"""
Video frame pipeline
====================

:mod:`araviq6.videostream` provides video pipeline classes to handle
``QVideoFrame`` using numpy array processing.

There are two options to handle the pipeline.

1. Frame-based approach
2. Array-based approach

The first approach can be achieved by connecting :class:`VideoFrameProcessor`
with ``QVideoSink``. The second is more low-level, and consists of
:class:`FrameToArrayConverter`, :class:`ArrayProcessor` and
:class:`ArrayToFrameConverter`

Convenience multimedia classes for array-based pipeline are also provided in this
module. These classes can replace the :class:`FrameToArrayConverter` connected
to ``QVideoSink``.

1. :class:`NDArrayVideoPlayer` (video file -> ndarray)
2. :class:`NDArrayMediaCaptureSession` (camera -> ndarray)

Pipeline classes
----------------

.. autoclass:: VideoFrameProcessor
   :members:
   :exclude-members: videoFrameProcessed

.. autoclass:: VideoFrameWorker
   :members:
   :exclude-members: videoFrameProcessed

.. autoclass:: QVideoFrameProperty
   :members:

.. autoclass:: FrameToArrayConverter
   :members:
   :exclude-members: arrayConverted

.. autoclass:: ArrayToFrameConverter
   :members:
   :exclude-members: frameConverted

.. autoclass:: ArrayProcessor
   :members:
   :exclude-members: arrayProcessed

.. autoclass:: ArrayWorker
   :members:
   :exclude-members: arrayProcessed

Convenience classes
-------------------

.. autoclass:: NDArrayVideoPlayer
   :members:
   :exclude-members: arrayChanged

.. autoclass:: NDArrayMediaCaptureSession
   :members:
   :exclude-members: arrayChanged

"""

import numpy as np
import numpy.typing as npt
import qimage2ndarray  # type: ignore[import]
from araviq6.array2qvideoframe import array2qvideoframe
from araviq6.qt_compat import QtCore, QtGui, QtMultimedia
from typing import Optional

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias


__all__ = [
    "VideoFrameWorker",
    "VideoFrameProcessor",
    "QVideoFrameProperty",
    "FrameToArrayConverter",
    "ArrayToFrameConverter",
    "ArrayWorker",
    "ArrayProcessor",
    "NDArrayVideoPlayer",
    "NDArrayMediaCaptureSession",
]


QVideoFrame: TypeAlias = QtMultimedia.QVideoFrame
MapMode: TypeAlias = QtMultimedia.QVideoFrame.MapMode
RotationAngle: TypeAlias = QtMultimedia.QVideoFrame.RotationAngle
QVideoFrameFormat: TypeAlias = QtMultimedia.QVideoFrameFormat


class VideoFrameWorker(QtCore.QObject):
    """
    Worker to process ``QVideoFrame`` using :class:`numpy.ndarray` operation.

    To perform processing, pass the input frame to :meth:`runProcess` and listen
    to :attr:`videoFrameProcessed` signal which emits two objects; processed
    QVideoFrame and processed NDArray.

    :meth:`ready` is set to ``False`` when the processing is running. This
    property can be utilized in multithreading.
    """

    videoFrameProcessed = QtCore.Signal(QVideoFrame, np.ndarray)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._ready = True

    def ready(self) -> bool:
        """
        Returns true if the worker finished processing and can process the next
        video frame without being blocked.
        """
        return self._ready

    def runProcess(self, frame: QVideoFrame):
        """
        Process *frame* and emit the results to :attr:`videoFrameProcessed`.

        When a video frame is passed, it is first converted to ``QImage`` by
        ``QVideoFrame.toImage()`` and then to array by :meth:`imageToArray`.
        Array processing is done by :meth:`processArray`, and the result is
        converted back to ``QVideoFrame`` by :meth:`arrayToVideoFrame`.

        Processed QVideoFrame and processed array are emitted by
        :attr:`videoFrameProcessed`.

        During the processing :meth:`ready` is set to False.

        Notes
        =====

        This method must *not* be Qt Slot to be multithreaded.
        """
        self._ready = False

        qimg = frame.toImage()  # must assign to avoid crash
        array = self.imageToArray(qimg)
        processedArray = self.processArray(array)
        processedFrame = self.arrayToVideoFrame(processedArray, frame)

        self.videoFrameProcessed.emit(processedFrame, processedArray)
        self._ready = True

    def imageToArray(self, image: QtGui.QImage) -> npt.NDArray[np.uint8]:
        """
        Convert *image* to numpy array.

        By default, this method uses ``qimage2ndarray.rgb_view`` for conversion.
        Null image is converted to 3D empty array. Subclass can redefine this
        method.
        """
        if image.isNull():
            ret = np.empty((0, 0, 0), dtype=np.uint8)
        else:
            ret = qimage2ndarray.rgb_view(image, byteorder=None)
        return ret

    def processArray(self, array: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
        """
        Perform image processing on *array* and return the result.

        By default this method does not perform any processing. Subclass can
        redefine this method.

        See also :meth:`runProcess`.
        """
        return array

    def arrayToVideoFrame(
        self, array: npt.NDArray[np.uint8], hintFrame: QVideoFrame
    ) -> QVideoFrame:
        """
        Convert *array* to ``QVideoFrame``, using *hintFrame* as hint.

        By default this method uses :func:`array2qvideoframe` for conversion.
        Then, it updates ``mapMode()``, ``startTime()``, and ``endTime()``
        properties of the new frame with those of *hintFrame*. For empty array,
        *hintFrame* is just returned. Subclass can redefine this method.
        """
        if array.size == 0:
            ret = hintFrame
        else:
            ret = array2qvideoframe(array)
            ret.map(hintFrame.mapMode())
            ret.setStartTime(hintFrame.startTime())
            ret.setEndTime(hintFrame.endTime())
        return ret


class VideoFrameProcessor(QtCore.QObject):
    """
    Video pipeline component to process ``QVideoFrame``.

    .. figure:: ../_images/frame-processor.jpg
       :align: center

       QVideoFrame processing structure

    :class:`VideoFrameProcessor` runs :class:`VideoFrameWorker` in an internal
    thread to process the incoming video frame. To perform processing, pass the
    input frame to :meth:`processVideoFrame` slot and listen to
    :attr:`videoFrameProcessed` signal which emits two objects; processed
    QVideoFrame and processed NDArray.

    :meth:`skipIfRunning` defines whether the incoming video frames should be
    skipped when the worker is running.

    """

    _processRequested = QtCore.Signal(QVideoFrame)
    videoFrameProcessed = QtCore.Signal(QVideoFrame, np.ndarray)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._worker = None
        self._skipIfRunning = True

        self._processorThread = QtCore.QThread()
        self._processorThread.start()

    def worker(self) -> Optional[VideoFrameWorker]:
        """
        Worker to process the video frame.

        See also :meth:`setWorker`.
        """
        return self._worker

    def setWorker(self, worker: Optional[VideoFrameWorker]):
        """
        Set *worker* as video frame processor.

        See also :meth:`worker`.
        """
        oldWorker = self.worker()
        if oldWorker is not None:
            self._processRequested.disconnect(oldWorker.runProcess)
            oldWorker.videoFrameProcessed.disconnect(self.videoFrameProcessed)
            oldWorker.videoFrameProcessed.disconnect(self.videoFrameProcessed)
        self._worker = worker
        if worker is not None:
            self._processRequested.connect(worker.runProcess)
            worker.videoFrameProcessed.connect(self.videoFrameProcessed)
            worker.videoFrameProcessed.connect(self.videoFrameProcessed)
            worker.moveToThread(self._processorThread)

    def skipIfRunning(self) -> bool:
        """
        If True, incoming frames to :meth:`processVideoFrame` are skipped if
        :meth:`worker` is running.

        If False, every incoming frame is queued if worker is not ready. This may
        consume a significant amount of memory and cause laggy displaying, thus
        should be used with discretion.

        See also :meth:`setSkipIfRunning`.
        """
        return self._skipIfRunning

    def setSkipIfRunning(self, flag: bool):
        """
        Set :meth:`skipIfRunning` to *flag*.

        See also :meth:`skipIfRunning`.
        """
        self._skipIfRunning = flag

    @QtCore.Slot(QVideoFrame)
    def processVideoFrame(self, frame: QVideoFrame):
        """
        Request :meth:`worker` to process *frame*.

        Processed QVideoFrame and processed array are emitted by
        :attr:`videoFrameProcessed`.

        If worker is running and :meth:`skipIfRunning` is True, *frame* is
        skipped without being emitted. If :meth:`skipIfRunning` is False,
        incoming frames are queued when worker is running.

        """
        worker = self.worker()
        if worker is not None:
            if worker.ready() or not self.skipIfRunning():
                self._processRequested.emit(frame)
        else:
            self.videoFrameProcessed.emit(frame)

    def stop(self):
        """Stop the worker thread."""
        self._processorThread.quit()
        self._processorThread.wait()


class QVideoFrameProperty:
    """Wrapper for the properties of QVideoFrame."""

    __slots__ = (
        "mapMode",
        "startTime",
        "endTime",
        "mirrored",
        "rotationAngle",
        "subtitleText",
    )

    def __init__(
        self,
        mapMode: MapMode = MapMode.NotMapped,
        startTime: int = -1,
        endTime: int = -1,
        mirrored: bool = False,
        rotationAngle: RotationAngle = RotationAngle.Rotation0,
        subtitleText: str = "",
    ):
        self.mapMode = mapMode
        self.startTime = startTime
        self.endTime = endTime
        self.mirrored = mirrored
        self.rotationAngle = rotationAngle
        self.subtitleText = subtitleText

    @classmethod
    def fromVideoFrame(cls, frame: QVideoFrame):
        """Construct :class:`QVideoFrameProperty` instance from *frame*."""
        return cls(*[getattr(frame, attr)() for attr in cls.__slots__])

    def setToVideoFrame(self, frame: QVideoFrame):
        """Set the properties of *frame* to the values in *self*."""
        frame.map(self.mapMode)
        frame.setStartTime(self.startTime)
        frame.setEndTime(self.endTime)
        frame.setMirrored(self.mirrored)
        frame.setRotationAngle(self.rotationAngle)
        frame.setSubtitleText(self.subtitleText)


class FrameToArrayConverter(QtCore.QObject):
    """
    Video pipeline component which converts ``QVideoFrame`` to numpy array.

    .. figure:: ../_images/frame-array-converter.jpg
       :align: center

       FrameToArrayConverter structure

    When video frame is passed to :meth:`convertVideoFrame`,
    :class:`FrameToArrayConverter` first converts it to ``QImage`` and then
    to numpy array using :meth:`imageToArray`. :attr:`arrayConverted` emits
    resulting array and :class:`QVideoFrameProperty`.

    Invalid video frame is converted to 3D empty array.

    """

    arrayConverted = QtCore.Signal(np.ndarray, QVideoFrameProperty)

    @QtCore.Slot(QVideoFrame)
    def convertVideoFrame(self, frame: QVideoFrame):
        """
        Convert *frame* to numpy array and emit to :attr:`arrayConverted`.

        Video frame is converted using using :meth:`imageToArray`, and its
        properties are wrapped by :class:`QVideoFrameProperty`. Converted array
        frame property are emitted to :meth:`arrayConverted`.

        """
        qimg = frame.toImage()
        if not qimg.isNull():
            array = self.imageToArray(qimg).copy()  # copy to detach reference
        else:
            array = np.empty((0, 0, 0), dtype=np.uint8)
        self.arrayConverted.emit(array, QVideoFrameProperty.fromVideoFrame(frame))

    def imageToArray(self, image: QtGui.QImage) -> npt.NDArray[np.uint8]:
        """
        Convert *image* to numpy array.

        *image* is ``QImage`` from ``QVideoFrame``. By default this method
        uses ``qimage2ndarray.rgb_view`` for conversion. Subclass can redefine
        this method.
        """
        return qimage2ndarray.rgb_view(image, byteorder=None)


class ArrayToFrameConverter(QtCore.QObject):
    """
    Video pipeline component which converts numpy array to ``QVideoFrame``.

    .. figure:: ../_images/array-frame-converter.jpg
       :align: center

       ArrayToFrameConverter structure

    Conversion is done by passing an array to :meth:`convertArray` slot and
    listening to :attr:`frameConverted` signal. Frame properties can be set by
    passing optional :class:`QVideoFrameProperty` with the array.

    """

    frameConverted = QtCore.Signal(QVideoFrame)

    @QtCore.Slot(np.ndarray, QVideoFrame)
    def convertArray(
        self,
        array: npt.NDArray[np.uint8],
        frameProperty: QVideoFrameProperty = QVideoFrameProperty(),
    ):
        """
        Convert *array* to ``QvideoFrame`` and emit to :attr:`frameConverted`.

        Valid array is converted using :meth:`arrayToFrame` with properties
        defined by *frameProperty*. Empty array is converted to invalid video
        frame.

        """
        if array.size != 0:
            frame = self.arrayToFrame(array)
        else:
            frameFormat = QVideoFrameFormat(
                QtCore.QSize(-1, -1), QVideoFrameFormat.PixelFormat.Format_Invalid
            )
            frame = QVideoFrame(frameFormat)
        frameProperty.setToVideoFrame(frame)
        self.frameConverted.emit(frame)

    def arrayToFrame(self, array: npt.NDArray[np.uint8]) -> QVideoFrame:
        """
        Converts *array* to ``QVideoFrame``.

        By default this method uses :func:`array2qvideoframe` for conversion.
        Subclass can redefine this method.
        """
        return array2qvideoframe(array)


class ArrayWorker(QtCore.QObject):
    """
    Worker to process numpy array.

    To perform processing, pass the input array to :meth:`runProcess` and
    listen to :attr:`arrayProcessed` signal.

    :meth:`ready` is set to ``False`` when the processing is being run. This
    property can be utilized in multithreading.
    """

    arrayProcessed = QtCore.Signal(np.ndarray)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._ready = True

    def ready(self) -> bool:
        """
        Returns true if the worker finished processing and can process the next
        array without being blocked.
        """
        return self._ready

    def runProcess(self, array: npt.NDArray[np.uint8]):
        """
        Process *array* and emit the result to :attr:`arrayProcessed`.

        Array processing is done by :meth:`processArray`.

        During the processing :meth:`ready` is set to False.

        Notes
        =====

        This method must *not* be Qt Slot to be multithreaded.
        """
        self._ready = False
        self.arrayProcessed.emit(self.processArray(array))
        self._ready = True

    def processArray(self, array: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
        """
        Perform image processing on *array* and return the result.

        By default this method does not perform any processing. Subclass can
        redefine this method.

        See also :meth:`runProcess`.
        """
        return array


class ArrayProcessor(QtCore.QObject):
    """
    Video pipeline component to process numpy array.

    .. figure:: ../_images/array-processor.jpg
       :align: center

       NDArray processing structure

    :class:`ArrayProcessor` runs :class:`ArrayWorker` in internal an thread t
    process the incoming array. To perform processing, pass the input array to
    :meth:`processArray` slot and listen to :attr:`arrayProcessed` signal.

    :meth:`skipIfRunning` defines whether the incoming arrays should be skipped
    when the worker is running.

    """

    _processRequested = QtCore.Signal(np.ndarray)
    arrayProcessed = QtCore.Signal(np.ndarray)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._worker = None
        self._skipIfRunning = True

        self._processorThread = QtCore.QThread()
        self._processorThread.start()

    def worker(self) -> Optional[ArrayWorker]:
        """
        Worker to process the array.

        See also :meth:`setWorker`.
        """
        return self._worker

    def setWorker(self, worker: Optional[ArrayWorker]):
        """
        Set *worker* as array processor.

        See also :meth:`worker`.
        """
        oldWorker = self.worker()
        if oldWorker is not None:
            self._processRequested.disconnect(oldWorker.runProcess)
            oldWorker.arrayProcessed.disconnect(self.arrayProcessed)
        self._worker = worker
        if worker is not None:
            self._processRequested.connect(worker.runProcess)
            worker.arrayProcessed.connect(self.arrayProcessed)
            worker.moveToThread(self._processorThread)

    def skipIfRunning(self) -> bool:
        """
        If True, incoming arrays to :meth:`processArray` are skipped if
        :meth:`worker` is not ready.

        If False, incoming arrays are queued if worker is not ready. This may
        consume a significant amount of memory and cause laggy displaying, thus
        should be used with discretion.

        See also :meth:`setSkipIfRunning`.
        """
        return self._skipIfRunning

    def setSkipIfRunning(self, flag: bool):
        """
        Set :meth:`skipIfRunning` to *flag*.

        See also :meth:`skipIfRunning`.
        """
        self._skipIfRunning = flag

    @QtCore.Slot(np.ndarray)
    def processArray(self, array: npt.NDArray[np.uint8]):
        """
        Request :meth:`worker` to process *array*.

        The result is emitted to :attr:`arrayProcessed`.

        If worker is running and :meth:`skipIfRunning` is True, *array* is
        skipped without being emitted. If :meth:`skipIfRunning` is False,
        incoming arrays are queued when worker is running.

        """
        worker = self.worker()
        if worker is not None:
            if worker.ready() or not self.skipIfRunning():
                self._processRequested.emit(array)
        else:
            self.arrayProcessed.emit(array)

    def stop(self):
        """Stop the worker thread."""
        self._processorThread.quit()
        self._processorThread.wait()


class NDArrayVideoPlayer(QtMultimedia.QMediaPlayer):
    """
    Video player which emits numpy array.

    When playing, :class:`NDArrayVideoPlayer` converts the frame to numpy array
    and :class:`QVideoFrameProperty`, and emits them to :attr:`arrayChanged`
    signal.

    User may use this class for convenience, or define own pipeline.
    """

    arrayChanged = QtCore.Signal(np.ndarray, QVideoFrameProperty)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._videoSink = QtMultimedia.QVideoSink()
        self._arrayConverter = FrameToArrayConverter(self)

        self.setVideoSink(self._videoSink)
        self._videoSink.videoFrameChanged.connect(
            self._arrayConverter.convertVideoFrame
        )
        self._arrayConverter.arrayConverted.connect(self.arrayChanged)


class NDArrayMediaCaptureSession(QtMultimedia.QMediaCaptureSession):
    """
    Capture session which emits numpy array.

    When capturing, :class:`NDArrayMediaCaptureSession` converts the frame to
    numpy array and :class:`QVideoFrameProperty`, and emits them to
    :attr:`arrayChanged` signal.

    User may use this class for convenience, or define own pipeline.
    """

    arrayChanged = QtCore.Signal(np.ndarray, QVideoFrameProperty)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._videoSink = QtMultimedia.QVideoSink()
        self._arrayConverter = FrameToArrayConverter(self)

        self.setVideoSink(self._videoSink)
        self._videoSink.videoFrameChanged.connect(
            self._arrayConverter.convertVideoFrame
        )
        self._arrayConverter.arrayConverted.connect(self.arrayChanged)



from araviq6 import get_samples_path, ClickableSlider
from araviq6.qt_compat import QtCore


VID_PATH = get_samples_path("hello.mp4")


def test_ClickableSlider(qtbot):
    slider = ClickableSlider()
    qtbot.addWidget(slider)
    assert slider.value() == 0

    pos = QtCore.QPoint(10, 10)
    qtbot.mouseClick(slider, QtCore.Qt.MouseButton.LeftButton, pos=pos)
    assert slider.value() == slider.pixelPosToRangeValue(pos)

    slider.close()



"""
General utilities
=================

:mod:`araviq6.util` provides utility classes and functions which are not
video pipeline components.

.. automodule:: araviq6.util.controller
   :members:

.. automodule:: araviq6.util.testing
   :members:

"""

from .controller import ClickableSlider, SignalProtocol, PlayerProtocol, MediaController
from .testing import get_samples_path


__all__ = [
    "ClickableSlider",
    "SignalProtocol",
    "PlayerProtocol",
    "MediaController",
    "get_samples_path",
]



"""
Media controller
================

:mod:`araviq6.util.controller` provides widget to control media player.

"""

from araviq6.qt_compat import QtCore, QtGui, QtMultimedia, QtWidgets
from typing import Optional

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[assignment]


__all__ = [
    "ClickableSlider",
    "SignalProtocol",
    "PlayerProtocol",
    "MediaController",
]


class ClickableSlider(QtWidgets.QSlider):
    """``QSlider`` whose groove can be clicked to move to position."""

    # https://stackoverflow.com/questions/52689047
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            val = self.pixelPosToRangeValue(event.position())
            self.setValue(val)
        super().mousePressEvent(event)

    def pixelPosToRangeValue(self, pos: QtCore.QPointF) -> int:
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        gr = self.style().subControlRect(
            QtWidgets.QStyle.ComplexControl.CC_Slider,
            opt,
            QtWidgets.QStyle.SubControl.SC_SliderGroove,
            self,
        )
        sr = self.style().subControlRect(
            QtWidgets.QStyle.ComplexControl.CC_Slider,
            opt,
            QtWidgets.QStyle.SubControl.SC_SliderHandle,
            self,
        )

        if self.orientation() == QtCore.Qt.Orientation.Horizontal:
            sliderLength = sr.width()
            sliderMin = gr.x()
            sliderMax = gr.right() - sliderLength + 1
        else:
            sliderLength = sr.height()
            sliderMin = gr.y()
            sliderMax = gr.bottom() - sliderLength + 1
        if self.orientation() == QtCore.Qt.Orientation.Horizontal:
            p = pos.x() - sr.center().x() + sr.topLeft().x()
        else:
            p = pos.y() - sr.center().y() + sr.topLeft().y()
        return QtWidgets.QStyle.sliderValueFromPosition(
            self.minimum(),
            self.maximum(),
            int(p - sliderMin),
            sliderMax - sliderMin,
            opt.upsideDown,  # type: ignore[attr-defined]
        )


class SignalProtocol(Protocol):
    def connect(
        self,
        receiver,
        type: QtCore.Qt.ConnectionType = QtCore.Qt.ConnectionType.AutoConnection,
    ):
        ...

    def disconnect(self, receiver):
        ...


class PlayerProtocol(Protocol):
    durationChanged: SignalProtocol
    positionChanged: SignalProtocol
    playbackStateChanged: SignalProtocol

    def playbackState(self) -> QtMultimedia.QMediaPlayer.PlaybackState:
        ...

    def play(self):
        ...

    def pause(self):
        ...

    def stop(self):
        ...

    def setPosition(self, position: int):
        ...


class MediaController(QtWidgets.QWidget):
    """
    Widget to control :class:`QtMultimedia.QMediaPlayer`.

    This controller can change the playback state and media position by
    :meth:`playButton`, :meth:`stopButton`, and :meth:`slider`.

    :meth:`setPlayer` sets the player to be controlled by this widget.

    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self._slider = ClickableSlider()
        self._playButton = QtWidgets.QPushButton()
        self._stopButton = QtWidgets.QPushButton()
        self._player = None
        self._pausedBySliderPress = False

        self._playButton.clicked.connect(self._onPlayButtonClick)
        self._stopButton.clicked.connect(self._onStopButtonClick)
        self._slider.sliderPressed.connect(self._onSliderPress)
        self._slider.sliderMoved.connect(self._onSliderMove)
        self._slider.sliderReleased.connect(self._onSliderRelease)

        layout = QtWidgets.QHBoxLayout()
        play_icon = self.style().standardIcon(
            QtWidgets.QStyle.StandardPixmap.SP_MediaPlay
        )
        self._playButton.setIcon(play_icon)
        layout.addWidget(self._playButton)
        stop_icon = self.style().standardIcon(
            QtWidgets.QStyle.StandardPixmap.SP_MediaStop
        )
        self._stopButton.setIcon(stop_icon)
        layout.addWidget(self._stopButton)
        self._slider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        layout.addWidget(self._slider)
        self.setLayout(layout)

    def player(self) -> Optional[PlayerProtocol]:
        """Media player which is controlled by *self*."""
        return self._player

    @QtCore.Slot()
    def _onPlayButtonClick(self):
        """Play or pause :meth:`player`."""
        player = self.player()
        if player is not None:
            if (
                player.playbackState()
                == QtMultimedia.QMediaPlayer.PlaybackState.PlayingState
            ):
                player.pause()
            else:
                player.play()

    @QtCore.Slot()
    def _onStopButtonClick(self):
        """Stop :meth:`player`."""
        player = self.player()
        if player is not None:
            player.stop()

    @QtCore.Slot()
    def _onSliderPress(self):
        """If the media was playing, pause and move to the pressed position."""
        player = self.player()
        if player is not None:
            if (
                player.playbackState()
                == QtMultimedia.QMediaPlayer.PlaybackState.PlayingState
            ):
                self._pausedBySliderPress = True
                player.pause()
            player.setPosition(self._slider.value())

    @QtCore.Slot(int)
    def _onSliderMove(self, position: int):
        """Move the media to current slider position."""
        player = self.player()
        if player is not None:
            player.setPosition(position)

    @QtCore.Slot()
    def _onSliderRelease(self):
        """If the media was paused by slider press, play the media."""
        player = self.player()
        if player is not None and self._pausedBySliderPress:
            player.play()
            self._pausedBySliderPress = False

    def setPlayer(self, player: Optional[PlayerProtocol]):
        """Set :meth:`player` and connect the signals."""
        old_player = self.player()
        if old_player is not None:
            old_player.durationChanged.disconnect(  # type: ignore[attr-defined]
                self._onMediaDurationChange
            )
            old_player.positionChanged.disconnect(  # type: ignore[attr-defined]
                self._onMediaPositionChange
            )
            old_player.playbackStateChanged.disconnect(  # type: ignore[attr-defined]
                self._onPlaybackStateChange
            )
        self._player = player
        if player is not None:
            player.durationChanged.connect(  # type: ignore[attr-defined]
                self._onMediaDurationChange
            )
            player.positionChanged.connect(  # type: ignore[attr-defined]
                self._onMediaPositionChange
            )
            player.playbackStateChanged.connect(  # type: ignore[attr-defined]
                self._onPlaybackStateChange
            )

    @QtCore.Slot("qint64")
    def _onMediaDurationChange(self, duration: int):
        """Set the slider range to media duration."""
        self._slider.setRange(0, duration)

    @QtCore.Slot("qint64")
    def _onMediaPositionChange(self, position: int):
        """Update the slider position to video position."""
        self._slider.setValue(position)

    @QtCore.Slot(QtMultimedia.QMediaPlayer.PlaybackState)
    def _onPlaybackStateChange(self, state: QtMultimedia.QMediaPlayer.PlaybackState):
        """Switch the play icon and pause icon by *state*."""
        if state == QtMultimedia.QMediaPlayer.PlaybackState.PlayingState:
            pause_icon = self.style().standardIcon(
                QtWidgets.QStyle.StandardPixmap.SP_MediaPause
            )
            self._playButton.setIcon(pause_icon)
        else:
            play_icon = self.style().standardIcon(
                QtWidgets.QStyle.StandardPixmap.SP_MediaPlay
            )
            self._playButton.setIcon(play_icon)



"""
Array label
===========

:mod:`araviq6.labels` provides ``QLabel`` subclasses to display numpy array.

.. autoclass:: ScalableQLabel
   :members:

.. autoclass:: NDArrayLabel
   :members:

"""

import enum
import numpy as np
import qimage2ndarray  # type: ignore[import]
from araviq6.qt_compat import QtCore, QtGui, QtWidgets

__all__ = [
    "ScalableQLabel",
    "NDArrayLabel",
]


class ScalableQLabel(QtWidgets.QLabel):
    """
    A label which can scale the pixmap before displaying.

    Pixmap can be downscaled or upscaled to fit to the label size, depending on
    :meth:`pixmapScaleMode`.

    :meth:`setPixmap` scales the input pixmap and update to label.
    :meth:`originalPixmap` returns current unscaled pixmap.

    Notes
    =====

    Do not modify the size policy and minimum size value. Changing them makes the
    label not shrinkable.

    """

    class PixmapScaleMode(enum.Enum):
        """
        This enum defines how the pixmap is scaled before being displayed.

        Attributes
        ==========

        PM_NoScale
            Pixmap is never scaled. If the label size is smaller than the pixmap
            size, only a part of the pixmap is displayed.

        PM_DownScaleOnly
            Pixmap is scaled, but never larger than its original size.

        PM_UpScaleOnly
            Pixmap is scaled, but never smaller than its original size.

        PM_AllScale
            Pixmap is scaled to any size.

        """

        PM_NoScale = 0
        PM_DownScaleOnly = 1
        PM_UpScaleOnly = 2
        PM_AllScale = 3

    PM_NoScale = PixmapScaleMode.PM_NoScale
    PM_DownScaleOnly = PixmapScaleMode.PM_DownScaleOnly
    PM_UpScaleOnly = PixmapScaleMode.PM_UpScaleOnly
    PM_AllScale = PixmapScaleMode.PM_AllScale

    def __init__(self, parent=None):
        super().__init__(parent)

        self._original_pixmap = QtGui.QPixmap()
        self._pixmapScaleMode = self.PM_DownScaleOnly
        # make label shrinkable
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding,
        )
        self.setMinimumSize(1, 1)  # (0, 0) prevents resizing

    def originalPixmap(self) -> QtGui.QPixmap:
        """Original pixmap before scaling."""
        return self._original_pixmap

    def pixmapScaleMode(self) -> PixmapScaleMode:
        """Mode to scale the pixmap. Default is :attr:`PM_DownScaleOnly`."""
        return self._pixmapScaleMode

    @QtCore.Slot(PixmapScaleMode)
    def setPixmapScaleMode(self, flag: PixmapScaleMode):
        """Set :meth:`pixmapScaleMode` to *flag* and update the label."""
        self._pixmapScaleMode = flag
        self.update()

    @QtCore.Slot(QtGui.QPixmap)
    def setPixmap(self, pixmap: QtGui.QPixmap):
        """Scale the pixmap and display."""
        self._original_pixmap = pixmap

        mode = self.pixmapScaleMode()
        if mode == self.PM_NoScale:
            flag = False
        else:
            w, h = pixmap.width(), pixmap.height()
            new_w = self.width()
            new_h = self.height()
            if mode == self.PM_DownScaleOnly:
                flag = new_w < w or new_h < h
            elif mode == self.PM_UpScaleOnly:
                flag = new_w > w or new_h > h
            elif mode == self.PM_AllScale:
                flag = True
            else:
                msg = "Unrecognized pixmap scale mode: %s" % mode
                raise TypeError(msg)

        if flag:
            pixmap = pixmap.scaled(
                new_w, new_h, QtCore.Qt.AspectRatioMode.KeepAspectRatio
            )

        super().setPixmap(pixmap)

    def paintEvent(self, event):
        super().paintEvent(event)
        self.setPixmap(self.originalPixmap())


class NDArrayLabel(ScalableQLabel):
    """
    Scalable label which can receive and display :class:`numpy.ndarray` image.
    Image array can be set by :meth:`setArray`.

    Examples
    ========

    >>> import cv2
    >>> from PySide6.QtWidgets import QApplication
    >>> import sys
    >>> from araviq6 import NDArrayLabel, get_samples_path
    >>> img = cv2.imread(get_samples_path('hello.jpg'))
    >>> def runGUI():
    ...     app = QApplication(sys.argv)
    ...     label = NDArrayLabel()
    ...     label.setArray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
    ...     label.show()
    ...     app.exec()
    ...     app.quit()
    >>> runGUI()  # doctest: +SKIP

    """

    @QtCore.Slot(np.ndarray)
    def setArray(self, array: np.ndarray):
        """
        Convert the array to pixmap using :func:`qimage2ndarray.array2qimage`,
        and display.
        """
        if array.size > 0:
            pixmap = QtGui.QPixmap.fromImage(qimage2ndarray.array2qimage(array))
        else:
            pixmap = QtGui.QPixmap()
        self.setPixmap(pixmap)



"""
Video widgets
=============

:mod:`araviq6.videowidgets` provides convenience widgets with pre-built video
pipelines.

.. autoclass:: PlayerProcessWidget
   :members:

.. autoclass:: CameraProcessWidget
   :members:

"""

from .videostream import VideoFrameWorker, VideoFrameProcessor
from .util import MediaController
from araviq6.qt_compat import QtCore, QtWidgets, QtMultimedia, QtMultimediaWidgets


__all__ = [
    "PlayerProcessWidget",
    "CameraProcessWidget",
]


class PlayerProcessWidget(QtWidgets.QWidget):
    """
    Widget to play video file with array processing.

    By default this widget does not perform any array processing. User may define
    own :class:`VideoFrameWorker` instance and set by :meth:`setWorker`.

    Examples
    ========

    .. tabs::

        .. code-tab:: python PySide6

            from PySide6.QtCore import QUrl
            from PySide6.QtWidgets import QApplication
            import sys
            from araviq6 import PlayerProcessWidget, VideoFrameWorker
            from araviq6.util import get_samples_path
            class FlipWorker(VideoFrameWorker):
                def processArray(self, array):
                    return array[::-1]
            def runGUI():
                app = QApplication(sys.argv)
                w = PlayerProcessWidget()
                w.setWorker(FlipWorker())
                w.setSource(QUrl.fromLocalFile(get_samples_path('hello.mp4')))
                w.show()
                app.exec()
                app.quit()
            runGUI() # doctest: +SKIP

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._videoPlayer = QtMultimedia.QMediaPlayer()
        self._playerVideoSink = QtMultimedia.QVideoSink()
        self._frameProcessor = VideoFrameProcessor()
        self._videoWidget = QtMultimediaWidgets.QVideoWidget()
        self._mediaController = MediaController()

        # set up the pipeline
        self._videoPlayer.setVideoSink(self._playerVideoSink)
        self._playerVideoSink.videoFrameChanged.connect(
            self._frameProcessor.processVideoFrame
        )
        self._frameProcessor.videoFrameProcessed.connect(self.displayVideoFrame)

        self._mediaController.setPlayer(self._videoPlayer)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self._videoWidget)
        layout.addWidget(self._mediaController)
        self.setLayout(layout)

    def videoPlayer(self) -> QtMultimedia.QMediaPlayer:
        return self._videoPlayer

    def frameProcessor(self) -> VideoFrameProcessor:
        return self._frameProcessor

    def setSource(self, url: QtCore.QUrl):
        self._videoPlayer.setSource(url)

    def setWorker(self, worker: VideoFrameWorker):
        self._frameProcessor.setWorker(worker)

    @QtCore.Slot(QtMultimedia.QVideoFrame)
    def displayVideoFrame(self, frame: QtMultimedia.QVideoFrame):
        self._videoWidget.videoSink().setVideoFrame(frame)

    def closeEvent(self, event):
        self._frameProcessor.stop()
        super().closeEvent(event)


class CameraProcessWidget(QtWidgets.QWidget):
    """
    Widget to stream camera with array processing.

    By default this widget does not perform any array processing. User may define
    own :class:`VideoFrameWorker` instance and set by :meth:`setWorker`.

    Examples
    ========

    .. tabs::

        .. code-tab:: python PySide6

            from PySide6.QtWidgets import QApplication
            from PySide6.QtMultimedia import QCamera
            import sys
            from araviq6 import VideoFrameWorker, CameraProcessWidget
            class FlipWorker(VideoFrameWorker):
                def processArray(self, array):
                    return array[::-1]
            def runGUI():
                app = QApplication(sys.argv)
                widget = CameraProcessWidget()
                widget.setWorker(FlipWorker())
                camera = QCamera()
                widget.setCamera(camera)
                camera.start()
                widget.show()
                app.exec()
                app.quit()
            runGUI() # doctest: +SKIP

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._captureSession = QtMultimedia.QMediaCaptureSession()
        self._cameraVideoSink = QtMultimedia.QVideoSink()
        self._frameProcessor = VideoFrameProcessor()
        self._videoWidget = QtMultimediaWidgets.QVideoWidget()

        # set up the pipeline
        self._captureSession.setVideoSink(self._cameraVideoSink)
        self._cameraVideoSink.videoFrameChanged.connect(
            self._frameProcessor.processVideoFrame
        )
        self._frameProcessor.videoFrameProcessed.connect(self.displayVideoFrame)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self._videoWidget)
        self.setLayout(layout)

    def captureSession(self) -> QtMultimedia.QMediaCaptureSession:
        return self._captureSession

    def frameProcessor(self) -> VideoFrameProcessor:
        return self._frameProcessor

    @QtCore.Slot(QtMultimedia.QCamera)
    def setCamera(self, camera: QtMultimedia.QCamera):
        self._captureSession.setCamera(camera)

    def setWorker(self, worker: VideoFrameWorker):
        self._frameProcessor.setWorker(worker)

    @QtCore.Slot(QtMultimedia.QVideoFrame)
    def displayVideoFrame(self, frame: QtMultimedia.QVideoFrame):
        self._videoWidget.videoSink().setVideoFrame(frame)

    def closeEvent(self, event):
        self._frameProcessor.stop()
        super().closeEvent(event)



from araviq6.qt_compat import QtCore, QtWidgets


def test_import_QtCore():
    assert hasattr(QtCore, "Qt")
    assert hasattr(QtCore, "Signal")
    assert hasattr(QtCore, "Slot")


def test_import_QtWidgets():
    assert hasattr(QtWidgets, "QWidget")



"""
AraViQ6: NDArray from QVideoFrame with Qt6
==========================================

AraViQ6 is a package to build video pipeline with :class:`numpy.ndarray` from
Qt6's ``QVideoFrame``.

"""

from .version import __version__  # noqa

from .array2qvideoframe import (
    array2qvideoframe,
)
from .videostream import (
    VideoFrameWorker,
    VideoFrameProcessor,
    QVideoFrameProperty,
    FrameToArrayConverter,
    ArrayToFrameConverter,
    ArrayWorker,
    ArrayProcessor,
    NDArrayVideoPlayer,
    NDArrayMediaCaptureSession,
)
from .labels import ScalableQLabel, NDArrayLabel
from .videowidgets import (
    PlayerProcessWidget,
    CameraProcessWidget,
)
from .util import (
    ClickableSlider,
    MediaController,
    get_samples_path,
)


__all__ = [
    "array2qvideoframe",
    "VideoFrameWorker",
    "VideoFrameProcessor",
    "QVideoFrameProperty",
    "FrameToArrayConverter",
    "ArrayToFrameConverter",
    "ArrayWorker",
    "ArrayProcessor",
    "NDArrayVideoPlayer",
    "NDArrayMediaCaptureSession",
    "ScalableQLabel",
    "NDArrayLabel",
    "PlayerProcessWidget",
    "CameraProcessWidget",
    "ClickableSlider",
    "MediaController",
    "get_samples_path",
]



# pytest-xvfb may fail by running this code in ubuntu.
# It's because for unknown reason, ``QVideoFrame.toImage()`` suffers memory issue
# when the video frame is manually constructed and modified.
# This issue happens only when any other test file is run before this one.
# Should the user encounter such error, run the test for this file separately.

import numpy as np
import qimage2ndarray  # type: ignore[import]
from araviq6 import (
    array2qvideoframe,
    QVideoFrameProperty,
    FrameToArrayConverter,
    ArrayToFrameConverter,
)
from araviq6.qt_compat import QtMultimedia


def test_array2qvideoframe(qtbot):
    bgr_array1 = np.array([[[1]]], dtype=np.uint8)
    frame1 = array2qvideoframe(bgr_array1)
    assert np.all(
        qimage2ndarray.byte_view(frame1.toImage())
        == np.array([[[1, 1, 1, 255]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.rgb_view(frame1.toImage()) == np.array([[[1]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.alpha_view(frame1.toImage())
        == np.array([[[255]]], dtype=np.uint8)
    )

    bgr_array2 = np.array([[[1, 2]]], dtype=np.uint8)
    frame2 = array2qvideoframe(bgr_array2)
    assert np.all(
        qimage2ndarray.byte_view(frame2.toImage())
        == np.array([[[1, 1, 1, 2]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.rgb_view(frame2.toImage()) == np.array([[[1]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.alpha_view(frame2.toImage()) == np.array([[[2]]], dtype=np.uint8)
    )

    bgr_array3 = np.array([[[1, 2, 3]]], dtype=np.uint8)
    frame3 = array2qvideoframe(bgr_array3)
    assert np.all(
        qimage2ndarray.byte_view(frame3.toImage())
        == np.array([[[1, 2, 3, 255]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.rgb_view(frame3.toImage())
        == np.array([[[3, 2, 1]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.alpha_view(frame3.toImage())
        == np.array([[[255]]], dtype=np.uint8)
    )

    bgr_array4 = np.array([[[1, 2, 3, 4]]], dtype=np.uint8)
    frame4 = array2qvideoframe(bgr_array4)
    assert np.all(
        qimage2ndarray.byte_view(frame4.toImage())
        == np.array([[[1, 2, 3, 4]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.rgb_view(frame4.toImage())
        == np.array([[[3, 2, 1]]], dtype=np.uint8)
    )
    assert np.all(
        qimage2ndarray.alpha_view(frame4.toImage()) == np.array([[[4]]], dtype=np.uint8)
    )


def test_FrameToArrayConverter(qtbot):
    frame = array2qvideoframe(np.array([[[1, 2, 3]]], dtype=np.uint8))
    converter = FrameToArrayConverter()

    with qtbot.waitSignal(
        converter.arrayConverted,
        check_params_cb=lambda array, prop: array.size != 0
        and prop.startTime == frame.startTime()
        and prop.endTime == frame.endTime(),
    ):
        converter.convertVideoFrame(frame)

    with qtbot.waitSignal(
        converter.arrayConverted, check_params_cb=lambda array, _: array.size == 0
    ):
        converter.convertVideoFrame(QtMultimedia.QVideoFrame())


def test_ArrayToFrameConverter(qtbot):
    converter = ArrayToFrameConverter()

    prop = QVideoFrameProperty(startTime=10, endTime=20)
    with qtbot.waitSignal(
        converter.frameConverted,
        check_params_cb=lambda frame: frame.isValid()
        and frame.startTime() == prop.startTime
        and frame.endTime() == prop.endTime,
    ):
        converter.convertArray(np.array([[[1, 2, 3]]], dtype=np.uint8), prop)

    nullProp = QVideoFrameProperty()
    with qtbot.waitSignal(
        converter.frameConverted,
        check_params_cb=lambda frame: frame.isValid()
        and frame.startTime() == nullProp.startTime
        and frame.endTime() == nullProp.endTime,
    ):
        converter.convertArray(np.array([[[1, 2, 3]]], dtype=np.uint8))

    with qtbot.waitSignal(
        converter.frameConverted, check_params_cb=lambda frame: not frame.isValid()
    ):
        converter.convertArray(np.empty((0, 0, 0), dtype=np.uint8))



"""
Array-frame conversion
======================

:mod:`araviq6.array2qvideoframe` provides functions to convert numpy array to
``QVideoFrame``.

To convert ``QVideoFrame`` to numpy array, convert the frame to ``QImage`` by
``QVideoFrame.toImage()`` and use :mod:`qimage2ndarray` package.

.. note::
   This module imitates https://github.com/hmeine/qimage2ndarray.

.. autofunction:: array2qvideoframe

"""

import numpy as np
import numpy.typing as npt
import sys
from qimage2ndarray import _normalize255  # type: ignore[import]
from araviq6.qt_compat import QtCore, QtMultimedia, get_frame_data
from typing import Optional, Union, Tuple


__all__ = [
    "array2qvideoframe",
]


class ArrayInterfaceAroundQVideoFrame(object):
    __slots__ = ("__qvideoframe", "__array_interface__")

    def __init__(self, frame: QtMultimedia.QVideoFrame, bytes_per_pixel: int):
        self.__qvideoframe = frame

        self.__array_interface__ = dict(
            shape=(frame.height(), frame.width()),
            typestr="|u%d" % bytes_per_pixel,
            data=get_frame_data(frame),
            strides=(frame.bytesPerLine(0), bytes_per_pixel),
            version=3,
        )


def qvideoframeview(frame: QtMultimedia.QVideoFrame) -> np.ndarray:
    pixelFormat = frame.surfaceFormat().pixelFormat()
    if pixelFormat == QtMultimedia.QVideoFrameFormat.PixelFormat.Format_BGRA8888:
        bits = 32
    elif pixelFormat == QtMultimedia.QVideoFrameFormat.PixelFormat.Format_BGRX8888:
        bits = 32
    else:
        raise TypeError(f"Invalid pixel format: {pixelFormat}")
    interface = ArrayInterfaceAroundQVideoFrame(frame, bits // 8)
    return np.asarray(interface)


def byte_view(
    frame: QtMultimedia.QVideoFrame, byteorder: Optional[str] = "little"
) -> npt.NDArray[np.uint8]:
    raw = qvideoframeview(frame)
    result = raw.view(np.uint8).reshape(raw.shape + (-1,))
    if byteorder is not None and byteorder != sys.byteorder:
        result = result[..., ::-1]
    return result


def rgb_view(
    frame: QtMultimedia.QVideoFrame, byteorder: Optional[str] = "big"
) -> npt.NDArray[np.uint8]:
    if byteorder is None:
        byteorder = sys.byteorder
    bytes = byte_view(frame, byteorder)

    if byteorder == "little":
        result = bytes[..., :3]  # strip A off BGRA
    else:
        result = bytes[..., 1:]  # strip A off ARGB
    return result


def alpha_view(frame: QtMultimedia.QVideoFrame) -> npt.NDArray[np.uint8]:
    bytes = byte_view(frame, byteorder=None)
    if sys.byteorder == "little":
        ret = bytes[..., 3]
    else:
        ret = bytes[..., 0]
    return ret


def array2qvideoframe(
    array: np.ndarray, normalize: Union[bool, int, Tuple[int, int]] = False
) -> QtMultimedia.QVideoFrame:
    """
    Convert a 2D or 3D numpy array into 32-bit ``QVideoFrame``.

    The dimensions of a 3D array are ``(width, height, channels)``, and the
    channels can be 1, 2, 3 or 4. 2D array with ``(width, height)`` dimension is
    converted to ``(width, height, 1)``.

    Number of the channels is interpreted as follows:

    ========= ===================
    #channels interpretation
    ========= ===================
            1 scalar/gray
            2 scalar/gray + alpha
            3 RGB
            4 RGB + alpha
    ========= ===================

    Note that the scalar data will be converted into gray RGB triples.

    The parameter *normalize* can be used to normalize an frame's value range
    to 0-255.

    If *normalize* = ``(nmin, nmax)``:
        Scale & clip frame values from ``nmin..nmax`` to ``0..255``

    If *normalize* = ``nmax``:
        Lets ``nmin`` default to zero, i.e. scale & clip the range ``0..nmax`` to
        ``0..255``
    If *normalize* = ``True``:
        Scale frame values to ``0..255``, except for boolean arays, where
        ``False`` and ``True`` are mapped to ``0`` and ``255``. Same as passing
        ``(gray.min(), gray.max())``

    If `array` contains masked values, the corresponding pixels will be
    transparent in the result. Thus, the result be of ``Format_BGRA8888`` if the
    input already contains an alhpa channel (i.e., has shape ``(H, W, 4)``) or
    if there are masked pixels, and ``Format_BGRX8888`` otherwise.

    """
    dim = np.ndim(array)
    if dim == 2:
        array = array[..., None]
    elif dim != 3:
        raise ValueError(
            f"invalid number of dimensions for array2qvideoframe (got {dim} dimensions)"
        )

    h, w, ch = array.shape
    if ch not in (1, 2, 3, 4):
        raise ValueError(
            f"invalid number of channels for array2qvideoframe: (got {ch} channels)"
        )

    hasAlpha = np.ma.is_masked(array) or ch in (2, 4)
    if hasAlpha:
        pixelFormat = QtMultimedia.QVideoFrameFormat.PixelFormat.Format_BGRA8888
    else:
        pixelFormat = QtMultimedia.QVideoFrameFormat.PixelFormat.Format_BGRX8888
    frameFormat = QtMultimedia.QVideoFrameFormat(QtCore.QSize(w, h), pixelFormat)
    frame = QtMultimedia.QVideoFrame(frameFormat)

    frame.map(QtMultimedia.QVideoFrame.MapMode.WriteOnly)

    array = _normalize255(array, normalize)
    rgb = rgb_view(frame)
    if ch >= 3:
        rgb[:] = array[..., :3]
    else:
        rgb[:] = array[..., :1]  # scalar data

    alpha = alpha_view(frame)
    if ch in (2, 4):
        alpha[:] = array[..., -1]
    else:
        alpha[:] = 255
    if np.ma.is_masked(array):
        alpha[:] *= np.logical_not(
            np.any(array.mask, axis=-1)  # type: ignore[attr-defined]
        )

    frame.unmap()  # save the modified memory to frame instance

    return frame



from araviq6 import ScalableQLabel, NDArrayLabel, get_samples_path
from araviq6.qt_compat import QtCore, QtGui
import cv2  # type: ignore[import]
import numpy as np
from qimage2ndarray import array2qimage  # type: ignore[import]


IMG_PATH = get_samples_path("hello.jpg")


def test_ScalableQLabel(qtbot):
    label = ScalableQLabel()
    qtbot.addWidget(label)

    img = cv2.cvtColor(cv2.imread(IMG_PATH), cv2.COLOR_BGR2RGBA)
    pixmap = QtGui.QPixmap.fromImage(array2qimage(img))
    pixmap_image = pixmap.toImage()
    img_h, img_w = pixmap_image.height(), pixmap_image.width()

    label.setPixmapScaleMode(label.PM_NoScale)
    # pixmap size is fixed
    label.setPixmap(pixmap)
    assert label.pixmap().size() == QtCore.QSize(img_w, img_h)
    assert label.pixmap().toImage() == pixmap_image
    # test downscaling : size fixed to original size
    new_w, new_h = (int(img_w / 2), int(img_h / 2))
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().size() == QtCore.QSize(img_w, img_h)
    assert label.pixmap().toImage() == pixmap_image
    # test upscaling : size fixed to original size
    new_w, new_h = (2 * img_w, 2 * img_h)
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().size() == QtCore.QSize(img_w, img_h)
    assert label.pixmap().toImage() == pixmap_image

    label.setPixmapScaleMode(label.PM_DownScaleOnly)
    # test downscaling
    new_w, new_h = (int(img_w / 2), int(img_h / 2))
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().width() < img_w and label.pixmap().height() < img_h
    # test upscaling : maximum size == original size
    new_w, new_h = (2 * img_w, 2 * img_h)
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().size() == QtCore.QSize(img_w, img_h)
    assert label.pixmap().toImage() == pixmap_image

    label.setPixmapScaleMode(label.PM_UpScaleOnly)
    # test downscaling : minimum size == original size
    new_w, new_h = (int(img_w / 2), int(img_h / 2))
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().size() == QtCore.QSize(img_w, img_h)
    assert label.pixmap().toImage() == pixmap_image
    # test upscaling
    new_w, new_h = (2 * img_w, 2 * img_h)
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().width() > img_w and label.pixmap().height() > img_h

    label.setPixmapScaleMode(label.PM_AllScale)
    # test downscaling
    new_w, new_h = (int(img_w / 2), int(img_h / 2))
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().width() < img_w and label.pixmap().height() < img_h
    # test upscaling
    new_w, new_h = (2 * img_w, 2 * img_h)
    label.resize(new_w, new_h)
    label.setPixmap(pixmap)
    assert label.pixmap().width() > img_w and label.pixmap().height() > img_h

    label.close()


def test_NDArrayLabel(qtbot):
    label = NDArrayLabel()
    qtbot.addWidget(label)

    img = cv2.imread(IMG_PATH)
    h, w = img.shape[:2]

    label.setPixmapScaleMode(label.PM_NoScale)
    # pixmap size is fixed
    label.setArray(img)
    assert label.pixmap().size() == QtCore.QSize(w, h)
    # test downscaling : size fixed to original size
    new_w, new_h = (int(w / 2), int(h / 2))
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().size() == QtCore.QSize(w, h)
    # test upscaling : size fixed to original size
    new_w, new_h = (2 * w, 2 * h)
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().size() == QtCore.QSize(w, h)

    label.setPixmapScaleMode(label.PM_DownScaleOnly)
    # test downscaling
    new_w, new_h = (int(w / 2), int(h / 2))
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().width() < w and label.pixmap().height() < h
    # test upscaling : maximum size == original size
    new_w, new_h = (2 * w, 2 * h)
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().size() == QtCore.QSize(w, h)

    label.setPixmapScaleMode(label.PM_UpScaleOnly)
    # test downscaling : minimum size == original size
    new_w, new_h = (int(w / 2), int(h / 2))
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().size() == QtCore.QSize(w, h)
    # test upscaling : maximum size == original size
    new_w, new_h = (2 * w, 2 * h)
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().width() > w and label.pixmap().height() > h

    label.setPixmapScaleMode(label.PM_AllScale)
    # test downscaling
    new_w, new_h = (int(w / 2), int(h / 2))
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().width() < w and label.pixmap().height() < h
    # test upscaling
    new_w, new_h = (2 * w, 2 * h)
    label.resize(new_w, new_h)
    label.setArray(img)
    assert label.pixmap().width() > w and label.pixmap().height() > h

    label.close()


def test_NDArrayLabel_emptyimage(qtbot):
    label = NDArrayLabel()
    qtbot.addWidget(label)
    assert label.pixmap().isNull()

    empty_array = np.empty((0, 0, 0), dtype=np.uint8)
    label.setArray(empty_array)
    assert label.pixmap().isNull()

    label.close()



"""
Dynamic Qt import
=================

Qt API for PyQt6 and PySide6.

To access Qt objects independently to the installed Qt binding, import the Qt
subpackages from this module as if importing from the Qt binding package.

.. code:: python

    from araviq6.qt_compat import QtCore

Or you can use :obj:`qt_api` which is an instance of :class:`QtAPI`.

.. code:: python

    from araviq6.qt_compat import qt_api
    widget = qt_api.QtWidgets.QWidget()

Notes
-----

This module is not part of public API, therefore users must not rely on it.

Based on from https://github.com/pytest-dev/pytest-qt.

"""


import os


__all__ = [
    "QtAPI",
    "QtAPIError",
    "qt_api",
]


class QtAPI:
    """
    Interface to access the Qt binding package installed in the environment.

    The instance of this class can be treated as root namespace of Qt package.
    At construction, this object determines the Qt binding package where it will
    import the submodules. When attribute starting with ``Qt`` is called
    (e.g. ``QtCore`` - this is the naming rule for Qt subpackages) this object
    imports the module from root package.

    For example, the following code imports :mod:`PyQt6.QtWidgets` module where
    :class:`QWidget` is retrieved.

    .. code:: python

        from araviq6.qt_compat import QtAPI
        widget = QtAPI("PyQt6").QtWidgets.QWidget()

    Qt binding is determined by following steps.

    1. If *api* argument is passed, use it.

    2. If ``ARAVIQ_QT_API`` environment variable is set, use it.

    3. Try import the packages in following order.

        * PySide6
        * PyQt6

    Letter case does not matter when specifying the API. If importing fails,
    :class:`QtAPIError` is raised.

    """

    # When new API is supported, update README and intro.rst

    def __init__(self, api=os.environ.get("ARAVIQ_QT_API")):
        self._import_errors = {}

        def _can_import(name):
            try:
                __import__(name)
                return True
            except ModuleNotFoundError as e:
                self._import_errors[name] = str(e)
                return False

        # If api is specified, use it.
        if api is not None:
            qtapi = {
                "pyside6": "PySide6",
                "pyqt6": "PyQt6",
            }.get(api.lower())
            if qtapi is None:
                raise QtAPIError(f"Specified Qt API is not supported: '{api}'")
            try:
                __import__(f"{qtapi}.QtCore")
            except ModuleNotFoundError:
                raise QtAPIError(f"Specified Qt API is not installed: '{qtapi}'")
            self.qt_binding = qtapi
        # If api is not specified, try import supported Qt modules.
        # Not importing only the root namespace because when uninstalling from
        # conda, the namespace can still be there.
        elif _can_import("PySide6.QtCore"):
            self.qt_binding = "PySide6"
        elif _can_import("PyQt6.QtCore"):
            self.qt_binding = "PyQt6"
        else:
            errors = "\n".join(
                f"  {module}: {reason}"
                for module, reason in sorted(self._import_errors.items())
            )
            msg = "Supported Qt not installed.\n" + errors
            raise QtAPIError(msg)

        def _import_module(module_name):
            m = __import__(self.qt_binding, globals(), locals(), [module_name], 0)
            return getattr(m, module_name)

        self.QtCore = _import_module("QtCore")
        self.QtWidgets = _import_module("QtWidgets")
        self.QtGui = _import_module("QtGui")
        self.QtMultimedia = _import_module("QtMultimedia")
        self.QtMultimediaWidgets = _import_module("QtMultimediaWidgets")

        if self.qt_binding in ("PyQt6",):
            self.QtCore.Signal = self.QtCore.pyqtSignal
            self.QtCore.Slot = self.QtCore.pyqtSlot


class QtAPIError(Exception):
    pass


qt_api = QtAPI()

QtCore = qt_api.QtCore
QtWidgets = qt_api.QtWidgets
QtGui = qt_api.QtGui
QtMultimedia = qt_api.QtMultimedia
QtMultimediaWidgets = qt_api.QtMultimediaWidgets

binding = qt_api.qt_binding
if binding in ("PySide6",):

    def get_frame_data(frame: QtMultimedia.QVideoFrame):  # type: ignore[name-defined]
        return frame.bits(0)

elif binding in ("PyQt6",):

    def get_frame_data(frame: QtMultimedia.QVideoFrame):  # type: ignore[name-defined]
        return (int(frame.bits(0)), False)

else:
    raise QtAPIError(f"Specified Qt binding: '{binding}'")



from itertools import chain
import os
from setuptools import setup, find_packages  # type: ignore[import]


VERSION_FILE = "araviq6/version.py"


def get_version():
    with open(VERSION_FILE, "r") as f:
        exec(compile(f.read(), VERSION_FILE, "exec"))
    return locals()["__version__"]


def read_readme():
    with open("README.md", encoding="utf-8") as f:
        content = f.read()
    return content


def get_package_data():
    pkgname = "araviq6"
    pkg_datapaths = ["py.typed", "samples"]

    ret = []
    for path in pkg_datapaths:
        fullpath = os.path.join(pkgname, path)
        if os.path.isfile(fullpath):
            ret.append(path)
        else:
            for root, dirs, files in os.walk(fullpath):
                for filename in files:
                    filepath = os.path.join(root, filename)
                    ret.append(os.path.join(*filepath.split(os.sep)[1:]))

    return {pkgname: ret}


def read_requirements(path):
    with open(path, "r") as f:
        ret = f.read().splitlines()
    return ret


def get_extras_require():
    ret = {}
    ret["test"] = read_requirements("requirements/test.txt")
    ret["test-ci"] = read_requirements("requirements/test.txt") + read_requirements(
        "requirements/test-ci.txt"
    )
    ret["doc"] = read_requirements("requirements/doc.txt")
    ret["full"] = list(set(chain(*ret.values())))
    return ret


setup(
    name="araviq6",
    version=get_version(),
    python_requires=">=3.7",
    description="Package for converting QVideoFrame to NDArray with Qt6",
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: Software Development :: Widget Sets",
    ],
    author="Jisoo Song",
    author_email="jeesoo9595@snu.ac.kr",
    maintainer="Jisoo Song",
    maintainer_email="jeesoo9595@snu.ac.kr",
    url="https://github.com/JSS95/araviq6",
    license="MIT",
    packages=find_packages(),
    package_data=get_package_data(),
    install_requires=read_requirements("requirements/install.txt"),
    extras_require=get_extras_require(),
)



"""
Testing helpers
===============

:mod:`araviq6.util.testing` provides utilities to help writing unit tests.

"""

import os
import araviq6


__all__ = [
    "get_samples_path",
]


def get_samples_path(*paths: str) -> str:
    """
    Get the absolute path to the directory where the sample data are stored.

    Parameters
    ==========

    paths
        Subpaths under ``araviq6/samples/`` directory.

    Returns
    =======

    path
        Absolute path to the sample depending on the user's system.

    """
    module_path = os.path.abspath(araviq6.__file__)
    module_path = os.path.split(module_path)[0]
    sample_dir = os.path.join(module_path, "samples")
    sample_dir = os.path.normpath(sample_dir)
    sample_dir = os.path.normcase(sample_dir)

    path = os.path.join(sample_dir, *paths)
    return path
